@@ -1,6 +1,7 @@
 import logging
 import os
 import warnings
+from collections import defaultdict
 from multiprocessing.pool import ThreadPool
 
 import numpy as np
@@ -27,7 +28,7 @@ from pastml.visualisation.itol_manager import generate_itol_annotations
 from pastml.visualisation.tree_compressor import REASONABLE_NUMBER_OF_TIPS
 from pastml.visualisation import get_formatted_date
 
-PASTML_VERSION = '1.9.27'
+PASTML_VERSION = '1.9.29.3'
 
 warnings.filterwarnings("ignore", append=True)
 
@@ -123,10 +124,7 @@ def _serialize_acr(args):
         f.write('pastml_version\t{}\n'.format(PASTML_VERSION))
         for name in sorted(acr_result.keys()):
             if name not in [FREQUENCIES, STATES, MARGINAL_PROBABILITIES]:
-                if NUM_SCENARIOS == name:
-                    f.write('{}\t{:g}\n'.format(name, acr_result[name]))
-                else:
-                    f.write('{}\t{}\n'.format(name, acr_result[name]))
+                f.write('{}\t{}\n'.format(name, acr_result[name]))
         if is_ml(acr_result[METHOD]):
             for state, freq in zip(acr_result[STATES], acr_result[FREQUENCIES]):
                 f.write('{}\t{}\n'.format(state, freq))
@@ -145,7 +143,7 @@ def _serialize_acr(args):
 
 def reconstruct_ancestral_states(forest, character, states, prediction_method=MPPA, model=F81,
                                  params=None, reoptimise=False, avg_br_len=None, num_nodes=None, num_tips=None,
-                                 force_joint=True):
+                                 force_joint=True, tau=0):
     """
     Reconstructs ancestral states for the given character on the given tree.
 
@@ -161,6 +159,7 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
     :param model: (optional, default is F81) state evolution model to be used by PASTML.
     :type model: str
     :param prediction_method: (optional, default is MPPA) ancestral state prediction method to be used by PASTML.
+    :param prediction_method: (optional, default is MPPA) ancestral state prediction method to be used by PASTML.
     :type prediction_method: str
     :param num_nodes: (optional) total number of nodes in the given tree (including tips).
         If not specified, will be calculated.
@@ -178,6 +177,9 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
     :param reoptimise: (False by default) if set to True and the parameters are specified,
         they will be considered as an optimisation starting point instead, and the parameters will be optimised.
     :type reoptimise: bool
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
 
     :return: ACR result dictionary whose values depend on the prediction method.
     :rtype: dict
@@ -205,7 +207,7 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
         return ml_acr(forest=forest, character=character, prediction_method=prediction_method, model=model,
                       states=states,
                       avg_br_len=avg_br_len, num_nodes=num_nodes, num_tips=num_tips, freqs=freqs, sf=sf, kappa=kappa,
-                      force_joint=force_joint, reoptimise=reoptimise)
+                      force_joint=force_joint, reoptimise=reoptimise, tau=tau)
     if is_parsimonious(prediction_method):
         return parsimonious_acr(forest, character, prediction_method, states, num_nodes, num_tips)
 
@@ -214,7 +216,7 @@ def reconstruct_ancestral_states(forest, character, states, prediction_method=MP
 
 
 def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, force_joint=True, threads=0,
-        reoptimise=False):
+        reoptimise=False, tau=0):
     """
     Reconstructs ancestral states for the given tree and
     all the characters specified as columns of the given annotation dataframe.
@@ -244,6 +246,9 @@ def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, f
     :param force_joint: (optional, default is True) whether the JOINT state should be added to the MPPA prediction
         even when not selected by the Brier score
     :type force_joint: bool
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
 
     :param threads: (optional, default is 0, which stands for automatic) number of threads PastML can use for parallesation.
         By default detected automatically based on the system. Note that PastML will at most use as many threads
@@ -262,7 +267,9 @@ def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, f
     columns = preannotate_forest(df, forest)
     for i, tree in enumerate(forest):
         name_tree(tree, suffix='' if len(forest) == 1 else '_{}'.format(i))
-    collapse_zero_branches(forest, features_to_be_merged=df.columns)
+    tau = max(tau, 0)
+    if not tau:
+        collapse_zero_branches(forest, features_to_be_merged=df.columns)
 
     avg_br_len, num_nodes, num_tips = get_forest_stats(forest)
 
@@ -272,7 +279,7 @@ def acr(forest, df, prediction_method=MPPA, model=F81, column2parameters=None, f
 
     def _work(args):
         return reconstruct_ancestral_states(*args, avg_br_len=avg_br_len, num_nodes=num_nodes, num_tips=num_tips,
-                                            force_joint=force_joint)
+                                            force_joint=force_joint, tau=tau)
 
     prediction_methods = value2list(len(columns), prediction_method, MPPA)
     models = value2list(len(columns), model, F81)
@@ -322,10 +329,11 @@ def _quote(str_list):
 def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
                     columns=None, prediction_method=MPPA, model=F81, parameters=None,
                     name_column=None, root_date=None, timeline_type=TIMELINE_SAMPLED,
-                    tip_size_threshold=REASONABLE_NUMBER_OF_TIPS,
-                    out_data=None, html_compressed=None, html=None, work_dir=None,
+                    tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, colours=None,
+                    out_data=None, html_compressed=None, html=None, html_mixed=None, work_dir=None,
                     verbose=False, forced_joint=False, upload_to_itol=False, itol_id=None, itol_project=None,
-                    itol_tree_name=None, offline=False, threads=0, reoptimise=False):
+                    itol_tree_name=None, offline=False, threads=0, reoptimise=False, focus=None,
+                    resolve_polytomies=False, tau=0):
     """
     Applies PastML to the given tree(s) with the specified states and visualises the result (as html maps).
 
@@ -387,6 +395,9 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
     :param reoptimise: (False by default) if set to True and the parameters are specified,
         they will be considered as an optimisation starting point instead, and optimised.
     :type reoptimise: bool
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
 
     :param name_column: (optional) name of the annotation table column to be used for node names
         in the compressed map visualisation
@@ -400,6 +411,11 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         of size less than threshold-th largest tip from the compressed map (set to 1e10 to keep all).
         The larger it is the less tips will be trimmed.
     :type tip_size_threshold: int
+    :param focus: optional way to put a focus on certain character state values,
+        so that the nodes in these states are displayed
+        even if they do not pass the trimming threshold (tip_size_threshold argument).
+        Should be in the form character:state.
+    :type focus: str or list(str)
     :param timeline_type: (optional, by default is pastml.visualisation.cytoscape_manager.TIMELINE_SAMPLED)
         type of timeline visualisation: at each date/distance to root selected on the slider, either
         (pastml.visualisation.cytoscape_manager.TIMELINE_SAMPLED) all the lineages sampled after it are hidden; "
@@ -409,6 +425,23 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         after this date/distance to root are hidden, and the external branches are cut
         to the specified date/distance to root if needed;
     :type timeline_type: str
+    :param colours: optional way to specify the colours used for character state visualisation.
+        Could be specified as
+        (1a) a dict {column: {state: colour}},
+        where column corresponds to the character for which these parameters should be used,
+        or (1b) in a form {column: path_to_colour_file};
+        or (2) as a list of paths to colour files
+        (in the same order as ``columns`` argument that specifies characters)
+        possibly given only for the first few characters;
+        or (3) as a path to colour file (only for the first character).
+        Each file should be tab-delimited, with two columns: the first one containing character states,
+        and the second, named "colour", containing colours, in HEX format (e.g. #a6cee3).
+    :type colours: str or list(str) or dict
+    :param resolve_polytomies: (default False) when True, the polytomies with a state change
+        (i.e. a parent node, P, in state A has more than 2 children, including m > 1 children, C_1, ..., C_m, in state B)
+        are resolved by grouping together same-state (different from the parent state) nodes
+        (i.e. a new internal node N in state B is created and becomes the child of P and the parent of C_1, ..., C_m).
+    :type resolve_polytomies: bool
 
     :param out_data: path to the output annotation file with the reconstructed ancestral character states.
     :type out_data: str
@@ -416,6 +449,9 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
     :type html_compressed: str
     :param html: (optional) path to the output tree visualisation file (html).
     :type html: str
+    :param html_mixed: (optional) path to the output mostly compressed map visualisation file (html),
+        where the nodes in states specified with the focus argument are uncompressed.
+    :type html_mixed: str
     :param work_dir: (optional) path to the folder where pastml parameter, named tree
         and marginal probability (for marginal ML methods (pastml.ml.MPPA, pastml.ml.MAP) only) files are to be stored.
         Default is <path_to_input_file>/<input_file_name>_pastml. If the folder does not exist, it will be created.
@@ -483,7 +519,7 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
         threads = max(os.cpu_count(), 1)
 
     acr_results = acr(roots, df, prediction_method=prediction_method, model=model, column2parameters=parameters,
-                      force_joint=forced_joint, threads=threads, reoptimise=reoptimise)
+                      force_joint=forced_joint, threads=threads, reoptimise=reoptimise, tau=tau)
     column2states = {acr_result[CHARACTER]: acr_result[STATES] for acr_result in acr_results}
 
     if not out_data:
@@ -500,8 +536,8 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
 
     itol_result = None
     new_tree = os.path.join(work_dir, get_named_tree_file(tree))
-    features = list(column2states.keys())
-    nwks = [root.write(format_root_node=True, format=3, features=[DATE, DATE_CI] + features) for root in roots]
+    features = [DATE, DATE_CI] + list(column2states.keys())
+    nwks = [root.write(format_root_node=True, format=3, features=features) for root in roots]
     with open(new_tree, 'w+') as f:
         f.write('\n'.join(nwks))
     try:
@@ -520,6 +556,19 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
             'Did not manage to save the annotated tree in nexus format due to the following error: {}'.format(e))
         pass
 
+    if upload_to_itol or html or html_compressed:
+        if colours:
+            if isinstance(colours, str):
+                colours = [colours]
+            if isinstance(colours, list):
+                colours = dict(zip(df.columns, colours))
+            elif isinstance(colours, dict):
+                colours = {col_name2cat(col): cls for (col, cls) in colours.items()}
+            else:
+                raise ValueError('Colours should be either a list or a dict, got {}.'.format(type(colours)))
+        else:
+            colours = {}
+
     if threads > 1:
         pool = ThreadPool(processes=threads - 1)
         async_result = pool.map_async(func=_serialize_acr, iterable=((acr_res, work_dir) for acr_res in acr_results))
@@ -532,7 +581,7 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
                     pass
             itol_result = pool.apply_async(func=generate_itol_annotations,
                                            args=(column2states, work_dir, acr_results, state_df, age_label,
-                                                 new_tree, itol_id, itol_project, itol_tree_name))
+                                                 new_tree, itol_id, itol_project, itol_tree_name, colours))
     else:
         for acr_res in acr_results:
             _serialize_acr((acr_res, work_dir))
@@ -544,13 +593,53 @@ def pastml_pipeline(tree, data, data_sep='\t', id_index=0,
                 except:
                     pass
             generate_itol_annotations(column2states, work_dir, acr_results, state_df, age_label,
-                                      new_tree, itol_id, itol_project, itol_tree_name)
+                                      new_tree, itol_id, itol_project, itol_tree_name, colours)
 
-    if html or html_compressed:
+    if html or html_compressed or html_mixed:
         logger.debug('\n=============VISUALISATION=====================')
-        visualize(roots, column2states=column2states, html=html, html_compressed=html_compressed,
+
+        if (html_compressed or html_mixed) and focus:
+            def parse_col_val(cv):
+                cv = str(cv).strip()
+                colon_pos = cv.find(':')
+                if colon_pos == -1:
+                    if len(column2states) == 1 and cv in next(iter(column2states.values())):
+                        return next(iter(column2states.keys())), cv
+                    else:
+                        raise ValueError('Focus values should be in a form character:state, got {} instead.'.format(cv))
+                col, state = col_name2cat(cv[:colon_pos]), cv[colon_pos + 1:]
+                if col not in column2states:
+                    ml_col = get_personalized_feature_name(col, get_default_ml_method())
+                    if ml_col in column2states:
+                        col = ml_col
+                    else:
+                        mp_col = get_personalized_feature_name(col, get_default_mp_method())
+                        if mp_col in column2states:
+                            col = mp_col
+                        else:
+                            raise ValueError('Character {} specified for focus values is not found in metadata.'.format(
+                                cv[:colon_pos]))
+                if state not in column2states[col]:
+                    raise ValueError(
+                        'Character {} state {} not found among possible states in metadata.'.format(cv[:colon_pos],
+                                                                                                    state))
+                return col, state
+
+            if isinstance(focus, str):
+                focus = list(focus)
+            if not isinstance(focus, list):
+                raise ValueError(
+                    'Focus argument should be either a string or a list of strings, got {} instead.'.format(
+                        type(focus)))
+            focus_cv = [parse_col_val(_) for _ in focus]
+            focus = defaultdict(set)
+            for c, v in focus_cv:
+                focus[c].add(v)
+
+        visualize(roots, column2states=column2states, html=html, html_compressed=html_compressed, html_mixed=html_mixed,
                   name_column=name_column, tip_size_threshold=tip_size_threshold, date_label=age_label,
-                  timeline_type=timeline_type, work_dir=work_dir, local_css_js=offline)
+                  timeline_type=timeline_type, work_dir=work_dir, local_css_js=offline, column2colours=colours, focus=focus,
+                  should_resolve_polytomies=resolve_polytomies)
 
     if threads > 1:
         async_result.wait()
@@ -611,7 +700,6 @@ def _validate_input(columns, data, data_sep, root_dates, id_index, name_column,
         df = df[columns]
 
     df.columns = [col_name2cat(column) for column in df.columns]
-
     node_names = set.union(*[{n.name for n in root.traverse() if n.name} for root in roots])
     df_index_names = set(df.index)
     filtered_df = df.loc[node_names & df_index_names, :]
@@ -779,6 +867,9 @@ def main():
     acr_group.add_argument('--reoptimise', action='store_true',
                            help='if the parameters are specified, they will be considered as an optimisation '
                                 'starting point instead and optimised.')
+    acr_group.add_argument('--tau', required=False, type=float, default=0,
+                           help='A smoothing factor to apply to branch lengths during likelihood calculation. '
+                                'If set to zero (default), zero internal branches will be collapsed instead.')
 
     vis_group = parser.add_argument_group('visualisation-related arguments')
     vis_group.add_argument('-n', '--name_column', type=str, default=None,
@@ -811,6 +902,27 @@ def main():
                                 "in the folder specified by --work_dir, so that internet connection is not needed "
                                 "(but you must not move the output html files to any location "
                                 "other that the one specified by --html/--html_compressed).")
+    vis_group.add_argument('--colours', type=str, nargs='*',
+                           help='optional way to specify the colours used for character state visualisation. '
+                                'Should be in the same order '
+                                'as the ancestral characters (see -c, --columns) '
+                                'for which the reconstruction is to be preformed. '
+                                'Could be given only for the first few characters. '
+                                'Each file should be tab-delimited, with two columns: '
+                                'the first one containing character states, '
+                                'and the second, named "colour", containing colours, in HEX format (e.g. #a6cee3).')
+    vis_group.add_argument('--focus', type=str, nargs='*',
+                           help='optional way to put a focus on certain character state values, '
+                                'so that the nodes in these states are displayed '
+                                'even if they do not pass the trimming threshold (--tip_size_threshold). '
+                                'Should be in the form character:state.')
+    vis_group.add_argument('--resolve_polytomies', action='store_true',
+                           help='When specified, the polytomies with a state change '
+                                '(i.e. a parent node, P, in state A has more than 2 children, '
+                                'including m > 1 children, C_1, ..., C_m, in state B) are resolved '
+                                'by grouping together same-state (different from the parent state) nodes '
+                                '(i.e. a new internal node N in state B is created and becomes the child of P '
+                                'and the parent of C_1, ..., C_m).')
 
     out_group = parser.add_argument_group('output-related arguments')
     out_group.add_argument('-o', '--out_data', required=False, type=str,
@@ -825,6 +937,9 @@ def main():
                            help="path to the output compressed map visualisation file (html).")
     out_group.add_argument('-l', '--html', required=False, default=None, type=str,
                            help="path to the output full tree visualisation file (html).")
+    out_group.add_argument('--html_mixed', required=False, default=None, type=str,
+                           help="path to the output mostly compressed map visualisation file (html), "
+                                "where the nodes in states specified with --focus are uncompressed.")
     out_group.add_argument('-v', '--verbose', action='store_true',
                            help="print information on the progress of the analysis (to console)")
 
