@@ -105,7 +105,7 @@ def get_pij_method(model=F81, frequencies=None, kappa=None):
         return lambda t: get_hky_pij(t, frequencies, kappa)
 
 
-def get_bottom_up_loglikelihood(tree, character, frequencies, sf, kappa=None, is_marginal=True, model=F81):
+def get_bottom_up_loglikelihood(tree, character, frequencies, sf, kappa=None, is_marginal=True, model=F81, tau=0):
     """
     Calculates the bottom-up loglikelihood for the given tree.
     The likelihood for each node is stored in the corresponding feature,
@@ -123,6 +123,9 @@ def get_bottom_up_loglikelihood(tree, character, frequencies, sf, kappa=None, is
     :type frequencies: numpy.array
     :param sf: scaling factor
     :type sf: float
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
     :return: log likelihood
     :rtype: float
     """
@@ -133,59 +136,53 @@ def get_bottom_up_loglikelihood(tree, character, frequencies, sf, kappa=None, is
 
     get_pij = get_pij_method(model, frequencies, kappa)
     for node in tree.traverse('postorder'):
-        likelihood_array = np.ones(len(frequencies), dtype=np.float64) * getattr(node, allowed_state_feature)
+        log_likelihood_array = np.log10(np.ones(len(frequencies), dtype=np.float64)
+                                        * getattr(node, allowed_state_feature))
         factors = 0
         for child in node.children:
-            child_likelihoods = get_pij(child.dist * sf) * getattr(child, lh_feature)
+            child_likelihoods = get_pij((child.dist + tau) * sf) * getattr(child, lh_feature)
             if is_marginal:
                 child_likelihoods = child_likelihoods.sum(axis=1)
             else:
                 child_states = child_likelihoods.argmax(axis=1)
                 child.add_feature(lh_joint_state_feature, child_states)
                 child_likelihoods = child_likelihoods.max(axis=1)
-
-            factors += rescale(child_likelihoods, fraction_of_limit=len(node.children))
-            likelihood_array *= child_likelihoods
-
-        if np.all(likelihood_array == 0):
-            return -np.inf
-
-        factors += rescale(likelihood_array, fraction_of_limit=len(node.up.children) if not node.is_root() else 1)
-        node.add_feature(lh_feature, likelihood_array)
+            child_likelihoods = np.maximum(child_likelihoods, 0)
+            log_likelihood_array += np.log10(child_likelihoods)
+            factors += rescale_log(log_likelihood_array)
+        node.add_feature(lh_feature, np.power(10, log_likelihood_array))
         node.add_feature(lh_sf_feature, factors + sum(getattr(_, lh_sf_feature) for _ in node.children))
     root_likelihoods = getattr(tree, lh_feature) * frequencies
     root_likelihoods = root_likelihoods.sum() if is_marginal else root_likelihoods.max()
     return np.log(root_likelihoods) - getattr(tree, lh_sf_feature) * np.log(10)
 
 
-def rescale(likelihood_array, fraction_of_limit):
+def rescale_log(loglikelihood_array):
     """
     Rescales the likelihood array if it gets too small/large, by multiplying it by a factor of 10.
-    :param fraction_of_limit: int, to be rescaled the min (max) non-zero likelihood value should be
-    smaller that MIN_VALUE / fraction_of_limit (larger than MAX_LIMIT / fraction_of_limit).
-    :param likelihood_array: numpy array containing the likelihood to be rescaled
+    :param loglikelihood_array: numpy array containing the loglikelihood to be rescaled
     :return: float, factor of 10 by which the likelihood array has been multiplies.
     """
 
-    max_limit = MAX_VALUE / fraction_of_limit
-    min_limit = MIN_VALUE / fraction_of_limit
+    max_limit = MAX_VALUE
+    min_limit = MIN_VALUE
 
-    min_lh_value = np.log10(np.min(likelihood_array[np.nonzero(likelihood_array)]))
-    max_lh_value = np.log10(np.max(likelihood_array[np.nonzero(likelihood_array)]))
+    non_zero_loglh_array = loglikelihood_array[loglikelihood_array > -np.inf]
+    min_lh_value = np.min(non_zero_loglh_array)
+    max_lh_value = np.max(non_zero_loglh_array)
 
     factors = 0
     if max_lh_value > max_limit:
         factors = max_limit - max_lh_value - 1
-        likelihood_array *= np.power(10, factors)
     elif min_lh_value < min_limit:
-        factors = min(-min_lh_value, max_limit - max_lh_value)
-        likelihood_array *= np.power(10, factors)
+        factors = min(min_limit - min_lh_value + 1, max_limit - max_lh_value - 1)
+    loglikelihood_array += factors
     return factors
 
 
 def optimize_likelihood_params(forest, character, frequencies, sf, kappa, avg_br_len, observed_frequencies,
                                optimise_sf=True, optimise_frequencies=True, optimise_kappa=True,
-                               model=F81):
+                               model=F81, tau=0):
     """
     Optimizes the likelihood parameters (state frequencies and scaling factor) for the given trees.
 
@@ -205,6 +202,9 @@ def optimize_likelihood_params(forest, character, frequencies, sf, kappa, avg_br
     :type optimise_sf: bool
     :param optimise_frequencies: whether the state frequencies need to be optimised
     :type optimise_frequencies: bool
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
     :return: optimized parameters and log likelihood: ((frequencies, scaling_factor), optimum)
     :rtype: tuple
     """
@@ -232,7 +232,7 @@ def optimize_likelihood_params(forest, character, frequencies, sf, kappa, avg_br
             return np.nan
         freqs, sf_val, kappa_val = get_real_params_from_optimised(ps)
         res = sum(get_bottom_up_loglikelihood(tree=tree, character=character, frequencies=freqs, sf=sf_val,
-                                              kappa=kappa_val, is_marginal=True, model=model) for tree in forest)
+                                              kappa=kappa_val, is_marginal=True, model=model, tau=tau) for tree in forest)
         return np.inf if pd.isnull(res) else -res
 
     x0_JC = np.hstack((frequencies[:-1] / frequencies[-1] if optimise_frequencies else [],
@@ -260,7 +260,7 @@ def optimize_likelihood_params(forest, character, frequencies, sf, kappa, avg_br
     return get_real_params_from_optimised(x0_JC if log_lh_JC >= log_lh_EFT else x0_EFT), best_log_lh
 
 
-def calculate_top_down_likelihood(tree, character, frequencies, sf, kappa=None, model=F81):
+def calculate_top_down_likelihood(tree, character, frequencies, sf, kappa=None, model=F81, tau=0):
     """
     Calculates the top-down likelihood for the given tree.
     The likelihood for each node is stored in the corresponding feature,
@@ -286,36 +286,41 @@ def calculate_top_down_likelihood(tree, character, frequencies, sf, kappa=None,
     :type tree: ete3.Tree
     :param frequencies: state frequencies
     :type frequencies: numpy.array
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
     :return: void, stores the node top-down likelihoods in the get_personalised_feature_name(feature, TD_LH) feature.
     """
 
-    lh_feature = get_personalized_feature_name(character, TD_LH)
-    lh_sf_feature = get_personalized_feature_name(character, TD_LH_SF)
+    td_lh_feature = get_personalized_feature_name(character, TD_LH)
+    td_lh_sf_feature = get_personalized_feature_name(character, TD_LH_SF)
     bu_lh_feature = get_personalized_feature_name(character, BU_LH)
     bu_lh_sf_feature = get_personalized_feature_name(character, BU_LH_SF)
 
     get_pij = get_pij_method(model, frequencies, kappa)
     for node in tree.traverse('preorder'):
         if node.is_root():
-            node.add_feature(lh_feature, np.ones(len(frequencies), np.float64))
-            node.add_feature(lh_sf_feature, 0)
+            node.add_feature(td_lh_feature, np.ones(len(frequencies), np.float64))
+            node.add_feature(td_lh_sf_feature, 0)
             continue
 
         parent = node.up
-        parent_bu_likelihood = getattr(parent, bu_lh_feature)
 
-        node_pjis = np.transpose(get_pij(node.dist * sf))
+        node_pjis = np.transpose(get_pij((node.dist + tau) * sf))
         node_contribution = getattr(node, bu_lh_feature).dot(node_pjis)
+        node_contribution[node_contribution <= 0] = 1
 
-        parent_likelihood = getattr(parent, lh_feature) * parent_bu_likelihood
-        parent_likelihood[np.nonzero(parent_likelihood)] /= node_contribution[np.nonzero(parent_likelihood)]
-        factors = getattr(parent, lh_sf_feature) + getattr(parent, bu_lh_sf_feature) - getattr(node, bu_lh_sf_feature)
+        parent_loglikelihood = np.log10(getattr(parent, td_lh_feature)) \
+                               + np.log10(getattr(parent, bu_lh_feature)) - np.log10(node_contribution)
+        factors = getattr(parent, td_lh_sf_feature) \
+                  + getattr(parent, bu_lh_sf_feature) - getattr(node, bu_lh_sf_feature)
+        factors += rescale_log(parent_loglikelihood)
+        parent_likelihood = np.power(10, parent_loglikelihood)
 
         td_likelihood = parent_likelihood.dot(node_pjis)
-        factors += rescale(td_likelihood, fraction_of_limit=len(node.children) if not node.is_leaf() else 1)
 
-        node.add_feature(lh_feature, td_likelihood)
-        node.add_feature(lh_sf_feature, factors)
+        node.add_feature(td_lh_feature, td_likelihood)
+        node.add_feature(td_lh_sf_feature, factors)
 
 
 def initialize_allowed_states(tree, feature, states):
@@ -456,10 +461,11 @@ def calculate_marginal_likelihoods(tree, feature, frequencies):
     allowed_state_feature = get_personalized_feature_name(feature, ALLOWED_STATES)
 
     for node in tree.traverse('preorder'):
-        likelihood = getattr(node, bu_lh_feature) * getattr(node, td_lh_feature) * frequencies \
-                     * getattr(node, allowed_state_feature)
-        node.add_feature(lh_feature, likelihood)
-        node.add_feature(lh_sf_feature, getattr(node, td_lh_sf_feature) + getattr(node, bu_lh_sf_feature))
+        loglikelihood = np.log10(getattr(node, bu_lh_feature)) + np.log10(getattr(node, td_lh_feature)) \
+                        + np.log10(frequencies * getattr(node, allowed_state_feature))
+        factors = rescale_log(loglikelihood)
+        node.add_feature(lh_feature, np.power(10, loglikelihood))
+        node.add_feature(lh_sf_feature, factors + getattr(node, td_lh_sf_feature) + getattr(node, bu_lh_sf_feature))
 
         node.del_feature(bu_lh_feature)
         node.del_feature(bu_lh_sf_feature)
@@ -637,7 +643,7 @@ def get_state2allowed_states(states, by_name=True):
 
 
 def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_nodes, num_tips, freqs=None, sf=None,
-           kappa=None, force_joint=True, reoptimise=False):
+           kappa=None, force_joint=True, reoptimise=False, tau=0):
     """
     Calculates ML states on the trees and stores them in the corresponding feature.
 
@@ -661,6 +667,9 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
         they will be considered as an optimisation starting point instead, and the parameters will be optimised.
     :type reoptimise: bool
     :return: mapping between reconstruction parameters and values
+    :param tau: a smoothing factor to apply to branch lengths during likelihood calculation.
+        If set to zero (default), zero internal branches will be collapsed instead.
+    :type tau: float
     :rtype: dict
     """
     n = len(states)
@@ -706,7 +715,8 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
 
     for tree in forest:
         initialize_allowed_states(tree, character, states)
-        alter_zero_tip_allowed_states(tree, character)
+        if not tau:
+            alter_zero_tip_allowed_states(tree, character)
     if sf and not reoptimise:
         optimise_sf = False
     else:
@@ -722,7 +732,13 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
         optimise_kappa = False
 
     likelihood = sum(get_bottom_up_loglikelihood(tree=tree, character=character, frequencies=frequencies, sf=sf,
-                                                 kappa=kappa, is_marginal=True, model=model) for tree in forest)
+                                                 kappa=kappa, is_marginal=True, model=model, tau=tau) for tree in forest)
+    if np.isnan(likelihood) or likelihood == -np.inf:
+        raise PastMLLikelihoodError('Failed to calculate the likelihood for your tree, '
+                                    'please check that you do not have contradicting {} states specified '
+                                    'for internal tree nodes, '
+                                    'and if not - submit a bug at https://github.com/evolbioinfo/pastml/issues'
+                                    .format(character))
     if not optimise_sf and not optimise_frequencies and not optimise_kappa:
         logger.debug('All the parameters are fixed for {}:{}{}{}{}.'
                      .format(character,
@@ -744,11 +760,17 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
                              '\n\tkappa:\t{:.6f}'.format(kappa) if HKY == model else '',
                              '\n\tlog likelihood:\t{:.6f}'.format(likelihood)))
         if optimise_sf:
-            (_, sf, _), likelihood = optimize_likelihood_params(forest=forest, character=character,
-                                                                frequencies=frequencies, sf=sf, kappa=kappa,
-                                                                optimise_frequencies=False, optimise_sf=optimise_sf,
-                                                                optimise_kappa=False, avg_br_len=avg_br_len,
-                                                                model=model, observed_frequencies=observed_frequencies)
+            (_, sf, _), likelihood = \
+                optimize_likelihood_params(forest=forest, character=character, frequencies=frequencies, sf=sf,
+                                           kappa=kappa, optimise_frequencies=False, optimise_sf=optimise_sf,
+                                           optimise_kappa=False, avg_br_len=avg_br_len, model=model,
+                                           observed_frequencies=observed_frequencies, tau=tau)
+            if np.any(np.isnan(likelihood) or likelihood == -np.inf):
+                raise PastMLLikelihoodError('Failed to optimise the likelihood for your tree, '
+                                            'please check that you do not have contradicting {} states specified '
+                                            'for internal tree nodes, '
+                                            'and if not - submit a bug at https://github.com/evolbioinfo/pastml/issues'
+                                            .format(character))
             if optimise_frequencies or optimise_kappa:
                 logger.debug('Pre-optimised SF for {}:{}{}.'
                              .format(character,
@@ -761,8 +783,13 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
                                            sf=sf, kappa=kappa, optimise_frequencies=optimise_frequencies,
                                            optimise_sf=optimise_sf,
                                            optimise_kappa=optimise_kappa, avg_br_len=avg_br_len, model=model,
-                                           observed_frequencies=observed_frequencies)
-
+                                           observed_frequencies=observed_frequencies, tau=tau)
+            if np.any(np.isnan(likelihood) or likelihood == -np.inf):
+                raise PastMLLikelihoodError('Failed to calculate the likelihood for your tree, '
+                                            'please check that you do not have contradicting {} states specified '
+                                            'for internal tree nodes, '
+                                            'and if not - submit a bug at https://github.com/evolbioinfo/pastml/issues'
+                                            .format(character))
         logger.debug('Optimised {} values:{}{}{}{}'
                      .format(character,
                              ''.join('\n\tfrequency of {}:\t{:.6f}'.format(state, frequencies[
@@ -794,12 +821,14 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
 
     def process_restricted_likelihood_and_states(method):
         for tree in forest:
-            alter_zero_tip_allowed_states(tree, character)
-        restricted_likelihood = sum(get_bottom_up_loglikelihood(tree=tree, character=character,
-                                                                frequencies=frequencies, sf=sf, kappa=kappa,
-                                                                is_marginal=True, model=model) for tree in forest)
+            if not tau:
+                alter_zero_tip_allowed_states(tree, character)
+        restricted_likelihood = \
+            sum(get_bottom_up_loglikelihood(tree=tree, character=character, frequencies=frequencies, sf=sf, kappa=kappa,
+                                            is_marginal=True, model=model, tau=tau) for tree in forest)
         for tree in forest:
-            unalter_zero_tip_allowed_states(tree, character, state2index)
+            if not tau:
+                unalter_zero_tip_allowed_states(tree, character, state2index)
         note_restricted_likelihood(method, restricted_likelihood)
         process_reconstructed_states(method)
 
@@ -810,12 +839,13 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
 
     if prediction_method != MAP:
         # Calculate joint restricted likelihood
-        restricted_likelihood = sum(get_bottom_up_loglikelihood(tree=tree, character=character,
-                                                                frequencies=frequencies, sf=sf, kappa=kappa,
-                                                                is_marginal=False, model=model) for tree in forest)
+        restricted_likelihood = \
+            sum(get_bottom_up_loglikelihood(tree=tree, character=character, frequencies=frequencies, sf=sf, kappa=kappa,
+                                            is_marginal=False, model=model, tau=tau) for tree in forest)
         note_restricted_likelihood(JOINT, restricted_likelihood)
         for tree in forest:
-            unalter_zero_tip_joint_states(tree, character, state2index)
+            if not tau:
+                unalter_zero_tip_joint_states(tree, character, state2index)
             choose_ancestral_states_joint(tree, character, states, frequencies)
         process_reconstructed_states(JOINT)
 
@@ -823,13 +853,15 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
         mps = []
         for tree in forest:
             initialize_allowed_states(tree, character, states)
-            alter_zero_tip_allowed_states(tree, character)
+            if not tau:
+                alter_zero_tip_allowed_states(tree, character)
             get_bottom_up_loglikelihood(tree=tree, character=character, frequencies=frequencies, sf=sf, kappa=kappa,
-                                        is_marginal=True, model=model)
-            calculate_top_down_likelihood(tree, character, frequencies, sf, kappa=kappa, model=model)
-            unalter_zero_tip_allowed_states(tree, character, state2index)
+                                        is_marginal=True, model=model, tau=tau)
+            calculate_top_down_likelihood(tree, character, frequencies, sf, kappa=kappa, model=model, tau=tau)
+            if not tau:
+                unalter_zero_tip_allowed_states(tree, character, state2index)
             calculate_marginal_likelihoods(tree, character, frequencies)
-            # check_marginal_likelihoods(tree, feature)
+            # check_marginal_likelihoods(tree, character)
             mps.append(convert_likelihoods_to_probabilities(tree, character, states))
 
             choose_ancestral_states_map(tree, character, states)
@@ -844,10 +876,11 @@ def ml_acr(forest, character, prediction_method, model, states, avg_br_len, num_
                 for pars_acr_res in pars_acr_results:
                     for tree in forest:
                         _parsimonious_states2allowed_states(tree, pars_acr_res[CHARACTER], character, state2index)
-                        alter_zero_tip_allowed_states(tree, character)
+                        if not tau:
+                            alter_zero_tip_allowed_states(tree, character)
                     restricted_likelihood = \
                         sum(get_bottom_up_loglikelihood(tree=tree, character=character, frequencies=frequencies,
-                                                        sf=sf, kappa=kappa, is_marginal=True, model=model)
+                                                        sf=sf, kappa=kappa, is_marginal=True, model=model, tau=tau)
                             for tree in forest)
                     note_restricted_likelihood(pars_acr_res[METHOD], restricted_likelihood)
 
@@ -885,3 +918,15 @@ def _parsimonious_states2allowed_states(tree, ps_feature, feature, state2index):
         for state in pars_states:
             allowed_states[state2index[state]] = 1
         node.add_feature(allowed_state_feature, allowed_states)
+
+
+class PastMLLikelihoodError(Exception):
+
+    def __init__(self, *args):
+        self.message = args[0] if args else None
+
+    def __str__(self):
+        if self.message:
+            return 'PastMLLikelihoodError, {}'.format(self.message)
+        else:
+            return 'PastMLLikelihoodError has been raised.'
