@@ -8,8 +8,7 @@ def get_formatted_date(node, dates_are_dates):
         try:
             date = numeric2datetime(date).strftime("%d %b %Y")
             if ci is not None:
-                ci[0] = numeric2datetime(ci[0]).strftime("%d %b %Y")
-                ci[1] = numeric2datetime(ci[1]).strftime("%d %b %Y")
+                ci = [numeric2datetime(ci[0]).strftime("%d %b %Y"), numeric2datetime(ci[1]).strftime("%d %b %Y")]
         except:
             pass
     return date if ci is None else '{} ({}-{})'.format(str(date), str(ci[0]), str(ci[1]))
