@@ -8,12 +8,14 @@ from shutil import copyfile
 import numpy as np
 from jinja2 import Environment, PackageLoader
 
+from pastml.file import get_pastml_colour_file
 from pastml import numeric2datetime
 from pastml.tree import DATE, DATE_CI
 from pastml.visualisation import get_formatted_date
-from pastml.visualisation.colour_generator import get_enough_colours, WHITE
+from pastml.visualisation.colour_generator import get_enough_colours, WHITE, parse_colours
 from pastml.visualisation.tree_compressor import NUM_TIPS_INSIDE, TIPS_INSIDE, TIPS_BELOW, \
-    REASONABLE_NUMBER_OF_TIPS, compress_tree, INTERNAL_NODES_INSIDE, ROOTS, IS_TIP, ROOT_DATES
+    REASONABLE_NUMBER_OF_TIPS, compress_tree, INTERNAL_NODES_INSIDE, ROOTS, IS_TIP, ROOT_DATES, IN_FOCUS, AROUND_FOCUS, \
+    UP_FOCUS, IS_POLYTOMY
 
 JS_LIST = ["https://pastml.pasteur.fr/static/js/jquery.min.js",
            "https://pastml.pasteur.fr/static/js/jquery.qtip.min.js",
@@ -55,11 +57,11 @@ FONT_SIZE = 'node_fontsize'
 
 MILESTONE = 'mile'
 
-
 DATE_LABEL = 'date'
 
 DIST_TO_ROOT_LABEL = 'dist. to root'
 
+
 def get_fake_node(n_id, x, y):
     attributes = {ID: n_id, 'fake': 1}
     return _get_node(attributes, position=(x, y))
@@ -96,20 +98,24 @@ def get_scaling_function(y_m, y_M, x_m, x_M):
 
 
 def set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size, transform_e_size, state,
-                                 root_names, root_dates, suffix=''):
-    tips_inside, internal_nodes_inside, roots = getattr(n, TIPS_INSIDE, []), \
-                                                getattr(n, INTERNAL_NODES_INSIDE, []), \
-                                                getattr(n, ROOTS, [])
+                                 root_names, root_dates, suffix='', is_mixed=False):
+    tips_inside, tips_below, internal_nodes_inside, roots = \
+        getattr(n, TIPS_INSIDE, []), getattr(n, TIPS_BELOW, []), getattr(n, INTERNAL_NODES_INSIDE, []), getattr(n,
+                                                                                                                ROOTS,
+                                                                                                                [])
 
     def get_min_max_str(values, default_value=0):
         min_v, max_v = (min(len(_) for _ in values), max(len(_) for _ in values)) \
             if values else (default_value, default_value)
         return ' {}'.format('{}-{}'.format(min_v, max_v) if min_v != max_v else min_v), min_v, max_v
 
-    tips_below_str, _, max_n_tips_below = get_min_max_str([list(_) for _ in roots])
+    tips_below_str, _, max_n_tips_below = get_min_max_str(tips_below)
     tips_inside_str, _, max_n_tips = get_min_max_str(tips_inside)
     internal_ns_inside_str, _, _ = get_min_max_str(internal_nodes_inside)
-    n.add_feature('{}{}'.format(NODE_NAME, suffix), '{}{}'.format(state, tips_inside_str))
+    n.add_feature('{}{}'.format(NODE_NAME, suffix),
+                  '{}{}'.format(state, tips_inside_str)
+                  if not is_mixed or (not getattr(n, IN_FOCUS, False) and not getattr(n, UP_FOCUS, False)) else
+                  '{}{}{}'.format(state, ':' if state else '', root_names[0]))
     size_factor = 2 if getattr(n, UNRESOLVED, False) else 1
     n.add_feature('{}{}'.format(NODE_SIZE, suffix),
                   (size_scaling(transform_size(max_n_tips)) if max_n_tips else int(MIN_NODE_SIZE / 1.5))
@@ -139,7 +145,7 @@ def set_cyto_features_tree(n, state):
 
 
 def _forest2json_compressed(forest, compressed_forest, columns, name_feature, get_date, milestones=None,
-                            dates_are_dates=True):
+                            dates_are_dates=True, is_mixed=False):
     e_size_scaling, font_scaling, size_scaling, transform_e_size, transform_size = \
         get_size_transformations(compressed_forest)
 
@@ -165,7 +171,6 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
 
     n2state = {}
 
-
     # Set the cytoscape features
     for compressed_tree in compressed_forest:
         for n in compressed_tree.traverse():
@@ -174,7 +179,8 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
             root_names = [_.name for _ in getattr(n, ROOTS)]
             root_dates = [get_formatted_date(_, dates_are_dates) for _ in getattr(n, ROOTS)]
             set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling,
-                                         transform_size, transform_e_size, state, root_names, root_dates)
+                                         transform_size, transform_e_size, state, root_names, root_dates,
+                                         is_mixed=is_mixed)
 
     # Calculate node coordinates
     min_size = 2 * min(min(getattr(_, NODE_SIZE) for _ in compressed_tree.traverse())
@@ -224,24 +230,29 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
                 suffix = '_{}'.format(i)
                 for n in nodes:
                     state = n2state[n]
-                    tips_inside, internal_nodes_inside, roots = getattr(n, TIPS_INSIDE, []), \
-                                                                getattr(n, INTERNAL_NODES_INSIDE, []), \
-                                                                getattr(n, ROOTS, [])
-                    tips_inside_i, internal_nodes_inside_i, roots_i = [], [], []
-                    for ti, ini, root in zip(tips_inside, internal_nodes_inside, roots):
+                    tips_inside, tips_below, internal_nodes_inside, roots = getattr(n, TIPS_INSIDE, []), \
+                                                                            getattr(n, TIPS_BELOW, []), \
+                                                                            getattr(n, INTERNAL_NODES_INSIDE, []), \
+                                                                            getattr(n, ROOTS, [])
+                    tips_inside_i, tips_below_i, internal_nodes_inside_i, roots_i = [], [], [], []
+                    for ti, tb, ini, root in zip(tips_inside, tips_below, internal_nodes_inside, roots):
                         if get_date(root) <= milestone:
                             roots_i.append(root)
 
                             ti = filter_by_date(ti, milestone)
+                            tb = [_ for _ in tb if getattr(_, DATE) <= milestone]
                             ini = filter_by_date(ini, milestone)
 
                             tips_inside_i.append(ti + [_ for _ in ini if _.is_leaf()])
+                            tips_below_i.append(tb)
                             internal_nodes_inside_i.append([_ for _ in ini if not _.is_leaf()])
-                    n.add_features(**{TIPS_INSIDE: tips_inside_i, INTERNAL_NODES_INSIDE: internal_nodes_inside_i,
+                    n.add_features(**{TIPS_INSIDE: tips_inside_i, TIPS_BELOW: tips_below_i,
+                                      INTERNAL_NODES_INSIDE: internal_nodes_inside_i,
                                       ROOTS: roots_i})
                     if roots_i:
                         n.add_feature(MILESTONE, i)
-                        root_names = [getattr(_, BRANCH_NAME) if getattr(_, DATE) > milestone else _.name for _ in roots_i]
+                        root_names = [getattr(_, BRANCH_NAME) if getattr(_, DATE) > milestone else _.name for _ in
+                                      roots_i]
                         root_dates = [getattr(_, DATE) for _ in roots_i]
                         if dates_are_dates:
                             try:
@@ -250,7 +261,7 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
                                 pass
                         set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size,
                                                      transform_e_size, state, root_names=root_names, suffix=suffix,
-                                                     root_dates=root_dates)
+                                                     root_dates=root_dates, is_mixed=is_mixed)
                         nodes_i.append(n)
                 nodes = nodes_i
 
@@ -274,7 +285,7 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
 
         for child in sorted(n.children, key=lambda _: node2id[_]):
             edge_attributes = {feature: getattr(child, feature) for feature in child.features
-                               if feature.startswith('edge_') or feature == MILESTONE}
+                               if feature.startswith('edge_') or feature == MILESTONE or feature == IS_POLYTOMY}
             source_name = n_id
             edges.append(get_edge(source_name, node2id[child], **edge_attributes))
 
@@ -284,10 +295,10 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
 
 def _forest2json(forest, columns, name_feature, get_date, milestones=None, timeline_type=TIMELINE_SAMPLED,
                  dates_are_dates=True):
-
     min_root_date = min(getattr(tree, DATE) for tree in forest)
     width = sum(len(tree) for tree in forest)
-    height_factor = 300 * width / max((max(getattr(_, DATE) for _ in tree) - min_root_date + tree.dist) for tree in forest)
+    height_factor = 300 * width / max(
+        (max(getattr(_, DATE) for _ in tree) - min_root_date + tree.dist) for tree in forest)
     zero_dist = min(min(min(_.dist for _ in tree.traverse() if _.dist > 0) for tree in forest), 300) * height_factor / 2
 
     # Calculate node coordinates
@@ -335,7 +346,8 @@ def _forest2json(forest, columns, name_feature, get_date, milestones=None, timel
                     if TIMELINE_LTT == timeline_type:
                         # if it is LTT also cut the branches if needed
                         if getattr(n, DATE) > milestone:
-                            n.add_feature('{}{}'.format(EDGE_NAME, suffix), np.round(milestone - getattr(n.up, DATE), 3))
+                            n.add_feature('{}{}'.format(EDGE_NAME, suffix),
+                                          np.round(milestone - getattr(n.up, DATE), 3))
                         else:
                             n.add_feature('{}{}'.format(EDGE_NAME, suffix), np.round(n.dist, 3))
 
@@ -370,12 +382,13 @@ def _forest2json(forest, columns, name_feature, get_date, milestones=None, timel
 
         for child in n.children:
             edge_attributes = {feature: getattr(child, feature) for feature in child.features
-                               if feature.startswith('edge_') or feature == MILESTONE}
+                               if feature.startswith('edge_') or feature == MILESTONE or feature == IS_POLYTOMY}
             source_name = n_id
             target_name = 'fake_node_{}'.format(node2id[child])
             nodes.append(get_fake_node(target_name, x=n2x[child], y=n2y[n]))
             edges.append(get_edge(source_name, target_name, fake=1,
-                                  **{k: v for (k, v) in edge_attributes.items() if EDGE_NAME not in k}))
+                                  **{k: v for (k, v) in edge_attributes.items()
+                                     if EDGE_NAME not in k and IS_POLYTOMY not in k}))
             source_name = target_name
             edges.append(get_edge(source_name, node2id[child], **edge_attributes))
 
@@ -417,7 +430,7 @@ def get_tooltip(n, columns):
 
 def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2colour, compressed_forest,
                            milestone_label, timeline_type, milestones, get_date, work_dir, local_css_js=False,
-                           milestone_labels=None):
+                           milestone_labels=None, is_mixed=False):
     """
     Converts a forest to an html representation using Cytoscape.js.
 
@@ -439,7 +452,8 @@ def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2c
     if compressed_forest is not None:
         json_dict, clazzes \
             = _forest2json_compressed(forest, compressed_forest, columns, name_feature=name_feature, get_date=get_date,
-                                      milestones=milestones, dates_are_dates=milestone_label == DATE_LABEL)
+                                      milestones=milestones, dates_are_dates=milestone_label == DATE_LABEL,
+                                      is_mixed=is_mixed)
     else:
         json_dict, clazzes \
             = _forest2json(forest, columns, name_feature=name_feature, get_date=get_date, milestones=milestones,
@@ -479,15 +493,17 @@ def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2c
         os.makedirs(os.path.join(work_dir, 'fonts'), exist_ok=True)
 
         template_dir = os.path.join(os.path.abspath(os.path.split(__file__)[0]), '..', 'templates')
-        for _ in sorted(glob(os.path.join(template_dir, 'js', '*.js'))):
+        for _ in sorted(glob(os.path.join(template_dir, 'js', '*.js*'))):
             cp = os.path.join(work_dir, 'js', os.path.split(_)[1])
             copyfile(_, cp)
-            js_list.append(cp)
+            if cp.endswith('.js'):
+                js_list.append(cp)
         css_list = []
-        for _ in glob(os.path.join(template_dir, 'css', '*.css')):
+        for _ in glob(os.path.join(template_dir, 'css', '*.css*')):
             cp = os.path.join(work_dir, 'css', os.path.split(_)[1])
             copyfile(_, cp)
-            css_list.append(cp)
+            if cp.endswith('.css'):
+                css_list.append(cp)
         for _ in glob(os.path.join(template_dir, 'fonts', '*.*')):
             cp = os.path.join(work_dir, 'fonts', os.path.split(_)[1])
             copyfile(_, cp)
@@ -527,38 +543,59 @@ def get_column_value_str(n, column, format_list=True, list_value=''):
     return ' or '.join(sorted(values)) if format_list or len(values) == 1 else list_value
 
 
-def visualize(forest, column2states, work_dir, name_column=None, html=None, html_compressed=None,
+def visualize(forest, column2states, work_dir, name_column=None, html=None, html_compressed=None, html_mixed=None,
               tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, date_label='Dist. to root', timeline_type=TIMELINE_SAMPLED,
-              local_css_js=False):
+              local_css_js=False, column2colours=None, focus=None, should_resolve_polytomies=False):
     one_column = next(iter(column2states.keys())) if len(column2states) == 1 else None
 
     name2colour = {}
     state2color = None
     for column, states in column2states.items():
         num_unique_values = len(states)
-        colours = get_enough_colours(num_unique_values)
+        if column2colours and column in column2colours:
+            colours = parse_colours(column2colours[column], states)
+        else:
+            colours = get_enough_colours(num_unique_values)
         for value, col in zip(states, colours):
             name2colour[value if one_column else '{}_{}'.format(column, value)] = col
-        logging.getLogger('pastml').debug('Mapped states to colours for {} as following: {} -> {}.'
-                                          .format(column, states, colours))
+        state2color = dict(zip(states, colours))
         # let ambiguous values be white
         if one_column is None:
             name2colour['{}_'.format(column)] = WHITE
         if column == name_column:
-            state2color = dict(zip(states, colours))
             for tree in forest:
                 for n in tree.traverse():
                     sts = getattr(n, column, set())
                     if len(sts) == 1 and not n.is_root() and getattr(n.up, column, set()) == sts:
                         n.add_feature('edge_color', state2color[next(iter(sts))])
+        out_colour_file = os.path.join(work_dir, get_pastml_colour_file(column))
+        # Not using DataFrames to speed up document writing
+        with open(out_colour_file, 'w+') as f:
+            f.write('state\tcolour\n')
+            for s in sorted(states):
+                f.write('{}\t{}\n'.format(s, state2color[s]))
+        logging.getLogger('pastml').debug('Mapped states to colours for {} as following: {} -> {}, '
+                                          'and serialized this mapping to {}.'
+                                          .format(column, states, colours, out_colour_file))
     for tree in forest:
+        nodes_in_focus = set()
         for node in tree.traverse():
             if node.is_leaf():
                 node.add_feature(IS_TIP, True)
             node.add_feature(BRANCH_NAME, '{}-{}'.format(node.up.name if not node.is_root() else '', node.name))
             for column in column2states.keys():
-                if len(getattr(node, column, set())) != 1:
+                col_state = getattr(node, column, set())
+                if len(col_state) != 1:
                     node.add_feature(UNRESOLVED, 1)
+                if focus and col_state & focus[column]:
+                    nodes_in_focus.add(node)
+        for node in nodes_in_focus:
+            node.add_feature(IN_FOCUS, True)
+            if not node.is_root() and node.up not in nodes_in_focus:
+                node.up.add_feature(UP_FOCUS, True)
+            for c in node.children:
+                if c not in nodes_in_focus:
+                    c.add_feature(AROUND_FOCUS, True)
 
     if TIMELINE_NODES == timeline_type:
         def get_date(node):
@@ -592,6 +629,8 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
     if milestone_labels is None:
         milestone_labels = ['{:g}'.format(_) for _ in milestones]
 
+    if should_resolve_polytomies:
+        resolve_polytomies(column2states, forest, name_column, state2color)
     if html:
         total_num_tips = sum(len(tree) for tree in forest)
         if total_num_tips > MAX_TIPS_FOR_FULL_TREE_VISUALISATION:
@@ -605,82 +644,106 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
                                    name_feature='name', compressed_forest=None, milestone_label=date_label,
                                    timeline_type=timeline_type, milestones=milestones, get_date=get_date,
                                    work_dir=work_dir, local_css_js=local_css_js, milestone_labels=milestone_labels)
+    if html_compressed and html_mixed:
+        forest_mixed = [tree.copy() for tree in forest]
+    else:
+        forest_mixed = forest
 
     if html_compressed:
-        resolve_polytomies(column2states, forest, name_column, state2color)
-        compressed_forest = [compress_tree(tree, columns=column2states.keys(), tip_size_threshold=tip_size_threshold)
-                             for tree in forest]
+        compressed_forest = [compress_tree(tree, columns=column2states.keys(), tip_size_threshold=tip_size_threshold,
+                                           mixed=False) for tree in forest]
+        milestone_labels, milestones = update_milestones(forest, date_label, milestone_labels, milestones,
+                                                         timeline_type)
 
-        # If we trimmed a few tips while compressing and they happened to be the oldest/newest ones,
-        # we should update the milestones accordingly.
-        first_date, last_date = np.inf, -np.inf
-        for tree in forest:
-            for _ in (tree.traverse() if timeline_type in [TIMELINE_LTT, TIMELINE_NODES] else tree):
-                first_date = min(first_date, getattr(_, DATE))
-                last_date = max(last_date, getattr(_, DATE))
-
-        milestones = [ms for ms in milestones if first_date <= ms <= last_date]
-        if milestones[0] > first_date:
-            milestones.insert(0, first_date)
-        if milestones[-1] < last_date:
-            milestones.append(last_date)
-
-        if DATE_LABEL == date_label:
-            try:
-                milestone_labels = [numeric2datetime(_).strftime("%d %b %Y") for _ in milestones]
-            except:
-                pass
-        if milestone_labels is None:
-            milestone_labels = ['{:g}'.format(_) for _ in milestones]
-
-        save_as_cytoscape_html(forest, html_compressed, column2states=column2states, name2colour=name2colour,
+        save_as_cytoscape_html(forest, html_compressed,
+                               column2states=column2states, name2colour=name2colour,
                                name_feature=name_column, compressed_forest=compressed_forest,
                                milestone_label=date_label, timeline_type=timeline_type,
                                milestones=milestones, get_date=get_date, work_dir=work_dir, local_css_js=local_css_js,
-                               milestone_labels=milestone_labels)
+                               milestone_labels=milestone_labels, is_mixed=False)
+
+    if html_mixed:
+        mixed_forest = [compress_tree(tree, columns=column2states.keys(), tip_size_threshold=tip_size_threshold,
+                                      mixed=True) for tree in forest_mixed]
+        milestone_labels, milestones = update_milestones(forest_mixed, date_label, milestone_labels, milestones,
+                                                         timeline_type)
+        save_as_cytoscape_html(forest_mixed, html_mixed,
+                               column2states=column2states, name2colour=name2colour,
+                               name_feature=name_column, compressed_forest=mixed_forest,
+                               milestone_label=date_label, timeline_type=timeline_type,
+                               milestones=milestones, get_date=get_date, work_dir=work_dir, local_css_js=local_css_js,
+                               milestone_labels=milestone_labels, is_mixed=True)
+
+
+def update_milestones(forest, date_label, milestone_labels, milestones, timeline_type):
+    # If we trimmed a few tips while compressing and they happened to be the oldest/newest ones,
+    # we should update the milestones accordingly.
+    first_date, last_date = np.inf, -np.inf
+    for tree in forest:
+        for _ in (tree.traverse() if timeline_type in [TIMELINE_LTT, TIMELINE_NODES] else tree):
+            first_date = min(first_date, getattr(_, DATE))
+            last_date = max(last_date, getattr(_, DATE))
+    milestones = [ms for ms in milestones if first_date <= ms <= last_date]
+    if milestones[0] > first_date:
+        milestones.insert(0, first_date)
+    if milestones[-1] < last_date:
+        milestones.append(last_date)
+    if DATE_LABEL == date_label:
+        try:
+            milestone_labels = [numeric2datetime(_).strftime("%d %b %Y") for _ in milestones]
+        except:
+            pass
+    if milestone_labels is None:
+        milestone_labels = ['{:g}'.format(_) for _ in milestones]
+    return milestone_labels, milestones
 
 
 def resolve_polytomies(column2states, forest, name_column, state2color):
     columns = sorted(column2states.keys())
 
+    col2state2i = {c: dict(zip(states, range(len(states)))) for (c, states) in column2states.items()}
+
     def get_prediction(n):
-        return tuple(tuple(sorted(getattr(n, c, set()))) for c in columns)
+        return '.'.join('-'.join(str(i) for i in sorted([col2state2i[c][_] for _ in getattr(n, c, set())])) for c in columns)
 
     for tree in forest:
         todo = [tree]
         while todo:
             n = todo.pop()
+            n_state = get_prediction(n)
             todo.extend(n.children)
             if len(n.children) > 2:
-                n_date = getattr(n, DATE)
-                n_ci = getattr(n, DATE_CI, None)
                 state2children = defaultdict(list)
                 for c in n.children:
                     state2children[get_prediction(c)].append(c)
-                if len(state2children) > 1:
-                    i = 0
-                    for children in state2children.values():
-                        if len(children) > 1:
-                            pol = n.add_child(dist=0, name='{}.polytomy_{}'.format(n.name, i))
-                            i += 1
-                            child = children[0]
-                            pol.add_feature(DATE, n_date)
-                            pol.add_feature(DATE_CI, n_ci)
-                            for c in columns:
-                                pol.add_feature(c, getattr(child, c))
-                            if hasattr(child, UNRESOLVED):
-                                pol.add_feature(UNRESOLVED, getattr(child, UNRESOLVED))
-                            if name_column is not None:
-                                sts = getattr(pol, name_column, set())
-                                if len(sts) == 1:
-                                    col = state2color[next(iter(sts))]
-                                    if getattr(n, name_column, set()) == sts:
-                                        pol.add_feature('edge_color', col)
-                                    for c in children:
-                                        c.add_feature('edge_color', col)
-                            pol.add_feature(BRANCH_NAME, '{}-{}'.format(n.name, pol.name))
-                            for c in children:
-                                c.add_feature(BRANCH_NAME, '{}-{}'.format(pol.name, c.name))
-                            for c in children:
-                                n.remove_child(c)
-                                pol.add_child(c)
+                for state, children in state2children.items():
+                    state_change = state != n_state
+                    if (state_change or getattr(n, UP_FOCUS, False)) and len(children) > 1:
+                        child = min(children, key=lambda _: _.dist)
+                        dist = child.dist if state_change else 0
+                        pol = n.add_child(dist=dist, name='{}.polytomy_{}'.format(n.name, state))
+                        pol.add_feature(IS_POLYTOMY, 1)
+                        pol.add_feature(DATE, getattr(child, DATE) if state_change else getattr(n, DATE))
+                        pol.add_feature(DATE_CI, getattr(child, DATE_CI, None) if state_change else getattr(n, DATE_CI, None))
+                        for c in columns:
+                            pol.add_feature(c, getattr(child, c))
+                        if hasattr(child, UNRESOLVED):
+                            pol.add_feature(UNRESOLVED, getattr(child, UNRESOLVED))
+                        if hasattr(child, IN_FOCUS):
+                            pol.add_feature(IN_FOCUS, getattr(child, IN_FOCUS))
+                        if hasattr(child, AROUND_FOCUS):
+                            pol.add_feature(AROUND_FOCUS, getattr(child, AROUND_FOCUS))
+                        if name_column is not None:
+                            sts = getattr(pol, name_column, set())
+                            if len(sts) == 1:
+                                col = state2color[next(iter(sts))]
+                                if getattr(n, name_column, set()) == sts:
+                                    pol.add_feature('edge_color', col)
+                                for c in children:
+                                    c.add_feature('edge_color', col)
+                        pol.add_feature(BRANCH_NAME, '{}-{}'.format(n.name, pol.name))
+                        for c in children:
+                            c.add_feature(BRANCH_NAME, '{}-{}'.format(pol.name, c.name))
+                        for c in children:
+                            n.remove_child(c)
+                            pol.add_child(c, dist=c.dist - dist)
